"""Pyflakes support."""

from pyflakes import checker

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract

m = checker.messages
CODES = {
    m.UnusedImport.message: "W0611",
    m.RedefinedWhileUnused.message: "W0404",
    m.RedefinedInListComp.message: "W0621",
    m.ImportShadowedByLoopVar.message: "W0621",
    m.ImportStarUsed.message: "W0401",
    m.ImportStarUsage.message: "W0401",
    m.UndefinedName.message: "E0602",
    m.DoctestSyntaxError.message: "W0511",
    m.UndefinedExport.message: "E0603",
    m.UndefinedLocal.message: "E0602",
    m.DuplicateArgument.message: "E1122",
    m.LateFutureImport.message: "W0410",
    m.UnusedVariable.message: "W0612",
    m.ReturnWithArgsInsideGenerator.message: "E0106",
    m.ReturnOutsideFunction.message: "E0104",
}


class Linter(Abstract):
    """Pyflakes runner."""

    name = "pyflakes"

    def run_check(self, context: RunContext):  # noqa
        """Check code with pyflakes."""
        params = context.get_params("pyflakes")
        builtins = params.get("builtins", "")
        if builtins:
            builtins = builtins.split(",")

        check = checker.Checker(context.ast, context.filename, builtins=builtins)
        for msg in check.messages:
            context.push(
                lnum=msg.lineno,
                col=msg.col + 1,
                text=msg.message % msg.message_args,
                number=CODES.get(msg.message, ""),
                source="pyflakes",
            )


#  pylama:ignore=E501,C0301



"""MyPy support.

TODO: Error codes
"""
from __future__ import annotations

from mypy import api

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """MyPy runner."""

    name = "mypy"

    def run_check(self, ctx: RunContext):
        """Check code with mypy."""
        # Support stdin
        args = [ctx.temp_filename, "--follow-imports=skip", "--show-column-numbers"]
        stdout, _, _ = api.run(args)  # noqa

        for line in stdout.splitlines():
            if not line:
                continue
            message = _MyPyMessage(line)
            if message.valid:
                ctx.push(
                    source="mypy",
                    lnum=message.line_num,
                    col=message.column,
                    text=message.text,
                    type=message.types.get(message.message_type.strip(), "W"),
                )


class _MyPyMessage:
    """Parser for a single MyPy output line."""

    types = {"error": "E", "warning": "W", "note": "N"}

    valid = False

    def __init__(self, line):
        self.filename = None
        self.line_num = None
        self.column = None

        try:
            result = line.split(":", maxsplit=4)
            self.filename, line_num_txt, column_txt, self.message_type, text = result
        except ValueError:
            return

        try:
            self.line_num = int(line_num_txt.strip())
            self.column = int(column_txt.strip())
        except ValueError:
            return

        self.text = text.strip()
        self.valid = True



"""pydocstyle support."""

from argparse import ArgumentParser

from pydocstyle import ConventionChecker as PyDocChecker
from pydocstyle.violations import conventions

from pylama.context import RunContext
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """Check pydocstyle errors."""

    name = "pydocstyle"

    @classmethod
    def add_args(cls, parser: ArgumentParser):
        """Add --max-complexity option."""
        parser.add_argument(
            "--pydocstyle-convention",
            choices=list(conventions.keys()),
            help="choose the basic list of checked errors by specifying an existing convention.",
        )

    def run_check(self, ctx: RunContext):  # noqa
        """Check code with pydocstyle."""
        params = ctx.get_params("pydocstyle")
        options = ctx.options
        if options and options.pydocstyle_convention:
            params.setdefault("convention", options.pydocstyle_convention)
        convention_codes = conventions.get(params.get("convention"))
        for err in PyDocChecker().check_source(
            ctx.source,
            ctx.filename,
            params.get("ignore_decorators"),
            params.get("ignore_inline_noqa", False),
        ):
            if convention_codes is None or err.code in convention_codes:
                ctx.push(
                    lnum=err.line,
                    text=err.short_desc,
                    type="D",
                    number=err.code,
                    source="pydocstyle",
                )



"""SCM hooks. Integration with git and mercurial."""

from __future__ import absolute_import

import sys
from configparser import ConfigParser  # noqa
from os import chmod
from os import path as op
from subprocess import PIPE, Popen
from typing import List, Tuple

from .config import parse_options, setup_logger
from .main import LOGGER, check_paths, display_errors


def run(command: str) -> Tuple[int, List[bytes], List[bytes]]:
    """Run a shell command."""
    with Popen(command.split(), stdout=PIPE, stderr=PIPE) as pipe:
        (stdout, stderr) = pipe.communicate()
        return (
            pipe.returncode,
            [line.strip() for line in stdout.splitlines()],
            [line.strip() for line in stderr.splitlines()],
        )


def git_hook(error=True):
    """Run pylama after git commit."""
    _, files_modified, _ = run("git diff-index --cached --name-only HEAD")

    options = parse_options()
    setup_logger(options)
    candidates = [f.decode("utf-8") for f in files_modified]
    if candidates:
        errors = check_paths(candidates, options)
        display_errors(errors, options)
        sys.exit(int(error and bool(errors)))


def hg_hook(_, repo, node=None, **kwargs):  # noqa
    """Run pylama after mercurial commit."""
    seen = set()
    candidates = []
    if len(repo):
        for rev in range(repo[node], len(repo)):
            for file_ in repo[rev].files():
                file_ = op.join(repo.root, file_)
                if file_ in seen or not op.exists(file_):
                    continue
                seen.add(file_)
                candidates.append(file_)

    options = parse_options()
    setup_logger(options)
    if candidates:
        errors = check_paths(candidates, options)
        display_errors(errors, options)
        sys.exit(int(bool(errors)))


def install_git(path):
    """Install hook in Git repository."""
    hook = op.join(path, "pre-commit")
    with open(hook, "w", encoding='utf-8') as target:
        target.write(
            """#!/usr/bin/env python
import sys
from pylama.hook import git_hook

if __name__ == '__main__':
    sys.exit(git_hook())
"""
        )
    chmod(hook, 484)


def install_hg(path):
    """Install hook in Mercurial repository."""
    hook = op.join(path, "hgrc")
    if not op.isfile(hook):
        open(hook, "w+", encoding='utf-8').close()

    cfgp = ConfigParser()
    with open(hook, "r", encoding='utf-8') as source:
        cfgp.read_file(source)

    if not cfgp.has_section("hooks"):
        cfgp.add_section("hooks")

    if not cfgp.has_option("hooks", "commit"):
        cfgp.set("hooks", "commit", "python:pylama.hooks.hg_hook")

    if not cfgp.has_option("hooks", "qrefresh"):
        cfgp.set("hooks", "qrefresh", "python:pylama.hooks.hg_hook")

    with open(hook, "w+", encoding='utf-8') as target:
        cfgp.write(target)


def install_hook(path):
    """Auto definition of SCM and hook installation."""
    is_git = op.join(path, ".git", "hooks")
    is_hg = op.join(path, ".hg")
    if op.exists(is_git):
        install_git(is_git)
        LOGGER.warning("Git hook has been installed.")

    elif op.exists(is_hg):
        install_hg(is_hg)
        LOGGER.warning("Mercurial hook has been installed.")

    else:
        LOGGER.error("VCS has not found. Check your path.")
        sys.exit(1)


# pylama:ignore=F0401,E1103,D210,F0001



"""Code complexity checking."""
from argparse import ArgumentError

from mccabe import McCabeChecker

from pylama.context import RunContext
from pylama.lint import ArgumentParser
from pylama.lint import LinterV2 as Abstract


class Linter(Abstract):
    """Run complexity checking."""

    name = "mccabe"

    @classmethod
    def add_args(cls, parser: ArgumentParser):
        """Add --max-complexity option."""
        try:
            parser.add_argument(
                "--max-complexity", default=10, type=int, help="Max complexity threshold"
            )
        except ArgumentError:
            pass

    def run_check(self, ctx: RunContext):
        """Run Mccabe code checker."""
        params = ctx.get_params("mccabe")
        options = ctx.options
        if options:
            params.setdefault("max-complexity", options.max_complexity)

        McCabeChecker.max_complexity = int(params.get("max-complexity", 10))
        McCabeChecker._error_tmpl = "%r is too complex (%d)"
        number = McCabeChecker._code
        for lineno, offset, text, _ in McCabeChecker(ctx.ast, ctx.filename).run():
            ctx.push(
                col=offset + 1,
                lnum=lineno,
                number=number,
                text=text,
                type="C",
                source="mccabe",
            )


#  pylama:ignore=W0212



"""Pylama's shell support."""

import sys
import warnings
from json import dumps
from os import path as op
from os import walk
from pathlib import Path
from typing import List, Optional

from .check_async import check_async
from .config import CURDIR, Namespace, parse_options, setup_logger
from .core import LOGGER, run
from .errors import Error
from .utils import read_stdin

DEFAULT_FORMAT = "{filename}:{lnum}:{col} [{etype}] {number} {message} [{source}]"
MESSAGE_FORMATS = {
    "pylint": "{filename}:{lnum}: [{etype}] {number} {message} [{source}]",
    "pycodestyle": "{filename}:{lnum}:{col} {number} {message} [{source}]",
    "parsable": DEFAULT_FORMAT,
}


def check_paths(
    paths: Optional[List[str]],
    options: Namespace,
    code: str = None,
    rootdir: Path = None,
) -> List[Error]:
    """Check the given paths.

    :param rootdir: Root directory (for making relative file paths)
    :param options: Parsed pylama options (from pylama.config.parse_options)
    """
    paths = paths or options.paths
    if not paths:
        return []

    if code is None:
        candidates = []
        for path in paths or options.paths:
            if not op.exists(path):
                continue

            if not op.isdir(path):
                candidates.append(op.abspath(path))

            for root, _, files in walk(path):
                candidates += [op.relpath(op.join(root, f), CURDIR) for f in files]
    else:
        candidates = [paths[0]]

    if not candidates:
        return []

    if rootdir is None:
        path = candidates[0]
        rootdir = Path(path if op.isdir(path) else op.dirname(path))

    candidates = [path for path in candidates if path.endswith(".py")]

    if options.concurrent:
        return check_async(candidates, code=code, options=options, rootdir=rootdir)

    errors = []
    for path in candidates:
        errors += run(path=path, code=code, rootdir=rootdir, options=options)

    return errors


def check_path(
    options: Namespace,
    rootdir: str = None,
    candidates: List[str] = None,
    code: str = None,  # noqa
) -> List[Error]:
    """Support legacy code."""
    warnings.warn(
        "pylama.main.check_path is depricated and will be removed in pylama 9",
        DeprecationWarning,
    )
    return check_paths(
        candidates,
        code=code,
        options=options,
        rootdir=rootdir and Path(rootdir) or None,
    )


def shell(args: List[str] = None, error: bool = True):
    """Endpoint for console.

    Parse a command arguments, configuration files and run a checkers.
    """
    if args is None:
        args = sys.argv[1:]

    options = parse_options(args)
    setup_logger(options)
    LOGGER.info(options)

    # Install VSC hook
    if options.hook:
        from .hook import install_hook  # noqa

        for path in options.paths:
            return install_hook(path)

    if options.from_stdin and not options.paths:
        LOGGER.error("--from-stdin requires a filename")
        return sys.exit(1)

    errors = check_paths(
        options.paths,
        code=read_stdin() if options.from_stdin else None,
        options=options,
        rootdir=CURDIR,
    )
    display_errors(errors, options)

    if error:
        sys.exit(int(bool(errors)))

    return errors


def display_errors(errors: List[Error], options: Namespace):
    """Format and display the given errors."""
    if options.format == "json":
        LOGGER.warning(dumps([err.to_dict() for err in errors]))

    else:
        pattern = MESSAGE_FORMATS.get(options.format, DEFAULT_FORMAT)
        for err in errors:
            LOGGER.warning(err.format(pattern))


if __name__ == "__main__":
    shell()

# pylama:ignore=F0001



"""Custom module loader."""
from __future__ import annotations

from argparse import ArgumentParser
from importlib import import_module
from pathlib import Path
from pkgutil import walk_packages
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

from pkg_resources import iter_entry_points

LINTERS: Dict[str, Type[LinterV2]] = {}

if TYPE_CHECKING:
    from pylama.context import RunContext


class LinterMeta(type):
    """Register linters."""

    def __new__(mcs, name, bases, params):
        """Register linters."""
        cls: Type[LinterV2] = super().__new__(mcs, name, bases, params)
        if cls.name is not None:
            LINTERS[cls.name] = cls
        return cls


class Linter(metaclass=LinterMeta):
    """Abstract class for linter plugin."""

    name: Optional[str] = None

    @classmethod
    def add_args(cls, _: ArgumentParser):
        """Add options from linters.

        The method has to be a classmethod.
        """

    def run(self, _path: str, **_meta) -> List[Dict[str, Any]]:  # noqa
        """Legacy method (support old extenstions)."""
        return []


class LinterV2(Linter):
    """A new linter class."""

    def run_check(self, ctx: RunContext):
        """Check code."""


# Import default linters
for _, pname, _ in walk_packages([str(Path(__file__).parent)]):  # type: ignore
    try:
        import_module(f"{__name__}.{pname}")
    except ImportError:
        pass

# Import installed linters
for entry in iter_entry_points("pylama.linter"):
    if entry.name not in LINTERS:
        try:
            LINTERS[entry.name] = entry.load()
        except ImportError:
            pass



"""Pylint integration to Pylama."""
import logging
from argparse import ArgumentParser
from os import environ
from pathlib import Path
from typing import Dict

from pylint.interfaces import CONFIDENCE_LEVELS
from pylint.lint import Run
from pylint.reporters import BaseReporter

from pylama.context import RunContext
from pylama.lint import LinterV2 as BaseLinter

HOME_RCFILE = Path(environ.get("HOME", "")) / ".pylintrc"


logger = logging.getLogger("pylama")


class Linter(BaseLinter):
    """Check code with Pylint."""

    name = "pylint"

    @classmethod
    def add_args(cls, parser: ArgumentParser):
        """Add --max-complexity option."""
        parser.add_argument(
            "--pylint-confidence",
            choices=[cc.name for cc in CONFIDENCE_LEVELS],
            help="Only show warnings with the listed confidence levels.",
        )

    def run_check(self, ctx: RunContext):
        """Pylint code checking."""
        logger.debug("Start pylint")
        params = ctx.get_params("pylint")
        options = ctx.options
        if options:
            params.setdefault("max_line_length", options.max_line_length)
            params.setdefault("confidence", options.pylint_confidence)

        params.setdefault("enable", ctx.select | ctx.get_filter("pylint", "select"))
        params.setdefault("disable", ctx.ignore | ctx.get_filter("pylint", "ignore"))

        class Reporter(BaseReporter):
            """Handle messages."""

            def _display(self, _):
                pass

            def handle_message(self, msg):
                msg_id = msg.msg_id
                ctx.push(
                    filtrate=False,
                    col=msg.column + 1,
                    lnum=msg.line,
                    number=msg_id,
                    text=msg.msg,
                    type=msg_id[0],
                    source="pylint",
                )

        logger.debug(params)

        reporter = Reporter()
        args = _Params(params).to_attrs()
        Run([ctx.temp_filename] + args, reporter=reporter, exit=False)


class _Params:
    """Store pylint params."""

    def __init__(self, params: Dict):
        attrs = {
            name.replace("_", "-"): self.prepare_value(value)
            for name, value in params.items()
            if value
        }
        if HOME_RCFILE.exists():
            attrs["rcfile"] = HOME_RCFILE.as_posix()

        self.attrs = attrs

    @staticmethod
    def prepare_value(value):
        """Prepare value to pylint."""
        if isinstance(value, (list, tuple, set)):
            return ",".join(value)

        if isinstance(value, bool):
            return "y" if value else "n"

        return str(value)

    def to_attrs(self):
        """Convert to argument list."""
        return [f"--{key}={value}" for key, value in self.attrs.items()]  # noqa

    def __str__(self):
        return " ".join(self.to_attrs())

    def __repr__(self):
        return f"<Pylint {self}>"


# pylama:ignore=W0403



"""Parse arguments from command line and configuration files."""
import fnmatch
import logging
import os
import re
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Collection, Dict, List, Optional, Set, Union

from pylama import LOGGER, __version__
from pylama.libs import inirama
from pylama.lint import LINTERS

#: A default checkers
DEFAULT_LINTERS = "pycodestyle", "pyflakes", "mccabe"

CURDIR = Path.cwd()
HOMECFG = Path.home() / ".pylama.ini"
CONFIG_FILES = "pylama.ini", "setup.cfg", "tox.ini", "pytest.ini"

# Setup a logger
LOGGER.propagate = False
STREAM = logging.StreamHandler(sys.stdout)
LOGGER.addHandler(STREAM)


class _Default:
    def __init__(self, value=None):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"<_Default [{self.value}]>"


def split_csp_str(val: Union[Collection[str], str]) -> Set[str]:
    """Split comma separated string into unique values, keeping their order."""
    if isinstance(val, str):
        val = val.strip().split(",")
    return set(x for x in val if x)


def prepare_sorter(val: Union[Collection[str], str]) -> Optional[Dict[str, int]]:
    """Parse sort value."""
    if val:
        types = split_csp_str(val)
        return dict((v, n) for n, v in enumerate(types, 1))

    return None


def parse_linters(linters: str) -> List[str]:
    """Initialize choosen linters."""
    return [name for name in split_csp_str(linters) if name in LINTERS]


def get_default_config_file(rootdir: Path = None) -> Optional[str]:
    """Search for configuration file."""
    if rootdir is None:
        return DEFAULT_CONFIG_FILE

    for filename in CONFIG_FILES:
        path = rootdir / filename
        if path.is_file() and os.access(path, os.R_OK):
            return path.as_posix()

    return None


DEFAULT_CONFIG_FILE = get_default_config_file(CURDIR)


def setup_parser() -> ArgumentParser:
    """Create and setup parser for command line."""
    parser = ArgumentParser(description="Code audit tool for python.")
    parser.add_argument(
        "paths",
        nargs="*",
        default=_Default([CURDIR.as_posix()]),
        help="Paths to files or directories for code check.",
    )
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)
    parser.add_argument("--verbose", "-v", action="store_true", help="Verbose mode.")
    parser.add_argument(
        "--options",
        "-o",
        default=DEFAULT_CONFIG_FILE,
        metavar="FILE",
        help=(
            "Specify configuration file. "
            f"Looks for {', '.join(CONFIG_FILES[:-1])}, or {CONFIG_FILES[-1]}"
            f" in the current directory (default: {DEFAULT_CONFIG_FILE})"
        ),
    )
    parser.add_argument(
        "--linters",
        "-l",
        default=_Default(",".join(DEFAULT_LINTERS)),
        type=parse_linters,
        help=(f"Select linters. (comma-separated). Choices are {','.join(s for s in LINTERS)}."),
    )
    parser.add_argument(
        "--from-stdin",
        action="store_true",
        help="Interpret the stdin as a python script, "
        "whose filename needs to be passed as the path argument.",
    )
    parser.add_argument(
        "--concurrent",
        "--async",
        action="store_true",
        help="Enable async mode. Useful for checking a lot of files. ",
    )

    parser.add_argument(
        "--format",
        "-f",
        default=_Default("pycodestyle"),
        choices=["pydocstyle", "pycodestyle", "pylint", "parsable", "json"],
        help="Choose output format.",
    )
    parser.add_argument(
        "--abspath",
        "-a",
        action="store_true",
        default=_Default(False),
        help="Use absolute paths in output.",
    )
    parser.add_argument(
        "--max-line-length",
        "-m",
        default=_Default(100),
        type=int,
        help="Maximum allowed line length",
    )
    parser.add_argument(
        "--select",
        "-s",
        default=_Default(""),
        type=split_csp_str,
        help="Select errors and warnings. (comma-separated list)",
    )
    parser.add_argument(
        "--ignore",
        "-i",
        default=_Default(""),
        type=split_csp_str,
        help="Ignore errors and warnings. (comma-separated)",
    )
    parser.add_argument(
        "--skip",
        default=_Default(""),
        type=lambda s: [re.compile(fnmatch.translate(p)) for p in s.split(",") if p],
        help="Skip files by masks (comma-separated, Ex. */messages.py)",
    )
    parser.add_argument(
        "--sort",
        default=_Default(),
        type=prepare_sorter,
        help="Sort result by error types. Ex. E,W,D",
    )
    parser.add_argument("--report", "-r", help="Send report to file [REPORT]")
    parser.add_argument("--hook", action="store_true", help="Install Git (Mercurial) hook.")

    for linter_type in LINTERS.values():
        linter_type.add_args(parser)

    return parser


def parse_options(  # noqa
    args: List[str] = None, config: bool = True, rootdir: Path = CURDIR, **overrides
) -> Namespace:
    """Parse options from command line and configuration files."""
    # Parse args from command string
    parser = setup_parser()
    actions = dict((a.dest, a) for a in parser._actions)  # pylint: disable=protected-access

    options = parser.parse_args(args or [])
    options.file_params = {}
    options.linters_params = {}

    # Compile options from ini
    if config:
        cfg = get_config(options.options, rootdir=rootdir)
        for opt, val in cfg.default.items():
            LOGGER.info("Find option %s (%s)", opt, val)
            passed_value = getattr(options, opt, _Default())
            if isinstance(passed_value, _Default):
                if opt == "paths":
                    val = val.split()
                if opt == "skip":
                    val = fix_pathname_sep(val)
                setattr(options, opt, _Default(val))

        # Parse file related options
        for name, opts in cfg.sections.items():

            if name == cfg.default_section:
                continue

            if name.startswith("pylama"):
                name = name[7:]

            if name in LINTERS:
                options.linters_params[name] = dict(opts)
                continue

            mask = re.compile(fnmatch.translate(fix_pathname_sep(name)))
            options.file_params[mask] = dict(opts)

    # Override options
    for opt, val in overrides.items():
        setattr(options, opt, process_value(actions, opt, val))

    # Postprocess options
    for name in options.__dict__:
        value = getattr(options, name)
        if isinstance(value, _Default):
            setattr(options, name, process_value(actions, name, value.value))

    if options.concurrent and "pylint" in options.linters:
        LOGGER.warning("Can't parse code asynchronously with pylint enabled.")
        options.concurrent = False

    return options


def process_value(actions: Dict, name: str, value: Any) -> Any:
    """Compile option value."""
    action = actions.get(name)
    if not action:
        return value

    if callable(action.type):
        return action.type(value)

    if action.const:
        return bool(int(value))

    return value


def get_config(ini_path: str = None, rootdir: Path = None) -> inirama.Namespace:
    """Load configuration from INI."""
    config = inirama.Namespace()
    config.default_section = "pylama"

    cfg_path = ini_path or get_default_config_file(rootdir)
    if not cfg_path and HOMECFG.exists():
        cfg_path = HOMECFG.as_posix()

    if cfg_path:
        LOGGER.info("Read config: %s", cfg_path)
        config.read(cfg_path)

    return config


def setup_logger(options: Namespace):
    """Do the logger setup with options."""
    LOGGER.setLevel(logging.INFO if options.verbose else logging.WARN)
    if options.report:
        LOGGER.removeHandler(STREAM)
        LOGGER.addHandler(logging.FileHandler(options.report, mode="w"))

    if options.options:
        LOGGER.info("Try to read configuration from: %r", options.options)


def fix_pathname_sep(val: str) -> str:
    """Fix pathnames for Win."""
    return val.replace(os.altsep or "\\", os.sep)


# pylama:ignore=W0212,D210,F0001



"""Pylama's core functionality.

Prepare params, check a modeline and run the checkers.
"""
import os.path as op
from pathlib import Path
from typing import List

from pylama.config import CURDIR, LOGGER, Namespace
from pylama.context import RunContext
from pylama.errors import Error, default_sorter, remove_duplicates
from pylama.lint import LINTERS, LinterV2


def run(
    path: str,
    code: str = None,
    rootdir: Path = CURDIR,
    options: Namespace = None,
) -> List[Error]:
    """Run code checkers with the given params.

    :param path: (str) A file's path.
    """
    path = op.relpath(path, rootdir)

    with RunContext(path, code, options) as ctx:
        if ctx.skip:
            LOGGER.info("Skip checking for path: %s", path)

        else:
            for lname in ctx.linters or LINTERS:
                linter_cls = LINTERS.get(lname)
                if not linter_cls:
                    continue
                linter = linter_cls()
                LOGGER.info("Run [%s] %s", lname, path)
                if isinstance(linter, LinterV2):
                    linter.run_check(ctx)
                else:
                    for err_info in linter.run(
                        ctx.temp_filename, code=ctx.source, params=ctx.get_params(lname)
                    ):
                        ctx.push(source=lname, **err_info)

    if not ctx.errors:
        return ctx.errors

    errors = list(remove_duplicates(ctx.errors))

    sorter = default_sorter
    if options and options.sort:
        sort = options.sort
        sorter = lambda err: (sort.get(err.etype, 999), err.lnum)

    return sorted(errors, key=sorter)


# pylama:ignore=R0912,D210,F0001
